"""Defines custom Rich styles and their corresponding names for cmd2.

This module provides a centralized and discoverable way to manage Rich styles
used within the cmd2 framework. It defines a StrEnum for style names and a
dictionary that maps these names to their default style objects.

**Notes**

Cmd2 uses Rich for its terminal output, and while this module defines a set of
cmd2-specific styles, it's important to understand that these aren't the only
styles that can appear. Components like Rich tracebacks and the rich-argparse
library, which cmd2 uses for its help output, also apply their own built-in
styles. Additionally, app developers may use other Rich objects that have
their own default styles.

For a complete theming experience, you can create a custom theme that includes
styles from Rich and rich-argparse. The `cmd2.rich_utils.set_theme()` function
automatically updates rich-argparse's styles with any custom styles provided in
your theme dictionary, so you don't have to modify them directly.

You can find Rich's default styles in the `rich.default_styles` module.
For rich-argparse, the style names are defined in the
`rich_argparse.RichHelpFormatter.styles` dictionary.

"""

import sys

from rich.style import (
    Style,
    StyleType,
)

if sys.version_info >= (3, 11):
    from enum import StrEnum
else:
    from backports.strenum import StrEnum

from .colors import Color


class Cmd2Style(StrEnum):
    """An enumeration of the names of custom Rich styles used in cmd2.

    Using this enum allows for autocompletion and prevents typos when
    referencing cmd2-specific styles.

    This StrEnum is tightly coupled with `DEFAULT_CMD2_STYLES`. Any name
    added here must have a corresponding style definition there.
    """

    COMMAND_LINE = "cmd2.example"  # Command line examples in help text
    ERROR = "cmd2.error"  # Error text (used by perror())
    HELP_HEADER = "cmd2.help.header"  # Help table header text
    HELP_LEADER = "cmd2.help.leader"  # Text right before the help tables are listed
    SUCCESS = "cmd2.success"  # Success text (used by psuccess())
    TABLE_BORDER = "cmd2.table_border"  # Applied to cmd2's table borders
    WARNING = "cmd2.warning"  # Warning text (used by pwarning())


# Default styles used by cmd2. Tightly coupled with the Cmd2Style enum.
DEFAULT_CMD2_STYLES: dict[str, StyleType] = {
    Cmd2Style.COMMAND_LINE: Style(color=Color.CYAN, bold=True),
    Cmd2Style.ERROR: Style(color=Color.BRIGHT_RED),
    Cmd2Style.HELP_HEADER: Style(color=Color.BRIGHT_GREEN, bold=True),
    Cmd2Style.HELP_LEADER: Style(color=Color.CYAN, bold=True),
    Cmd2Style.SUCCESS: Style(color=Color.GREEN),
    Cmd2Style.TABLE_BORDER: Style(color=Color.BRIGHT_GREEN),
    Cmd2Style.WARNING: Style(color=Color.BRIGHT_YELLOW),
}



"""Machinery for running and validating transcripts.

If the user wants to run a transcript (see docs/transcript.rst),
we need a mechanism to run each command in the transcript as
a unit test, comparing the expected output to the actual output.

This file contains the class necessary to make that work. This
class is used in cmd2.py::run_transcript_tests()
"""

import re
import unittest
from collections.abc import Iterator
from typing import (
    TYPE_CHECKING,
    Optional,
    TextIO,
    cast,
)

from . import string_utils as su
from . import utils

if TYPE_CHECKING:  # pragma: no cover
    from cmd2 import (
        Cmd,
    )


class Cmd2TestCase(unittest.TestCase):
    """A unittest class used for transcript testing.

    Subclass this, setting CmdApp, to make a unittest.TestCase class
    that will execute the commands in a transcript file and expect the
    results shown.

    See transcript_example.py
    """

    cmdapp: Optional['Cmd'] = None

    def setUp(self) -> None:
        """Instructions that will be executed before each test method."""
        if self.cmdapp:
            self._fetch_transcripts()

            # Trap stdout
            self._orig_stdout = self.cmdapp.stdout
            self.cmdapp.stdout = cast(TextIO, utils.StdSim(self.cmdapp.stdout))

    def tearDown(self) -> None:
        """Instructions that will be executed after each test method."""
        if self.cmdapp:
            # Restore stdout
            self.cmdapp.stdout = self._orig_stdout

    def runTest(self) -> None:  # was testall  # noqa: N802
        """Override of the default runTest method for the unittest.TestCase class."""
        if self.cmdapp:
            its = sorted(self.transcripts.items())
            for fname, transcript in its:
                self._test_transcript(fname, transcript)

    def _fetch_transcripts(self) -> None:
        self.transcripts = {}
        testfiles = cast(list[str], getattr(self.cmdapp, 'testfiles', []))
        for fname in testfiles:
            with open(fname) as tfile:
                self.transcripts[fname] = iter(tfile.readlines())

    def _test_transcript(self, fname: str, transcript: Iterator[str]) -> None:
        if self.cmdapp is None:
            return

        line_num = 0
        finished = False
        line = su.strip_style(next(transcript))
        line_num += 1
        while not finished:
            # Scroll forward to where actual commands begin
            while not line.startswith(self.cmdapp.visible_prompt):
                try:
                    line = su.strip_style(next(transcript))
                except StopIteration:
                    finished = True
                    break
                line_num += 1
            command_parts = [line[len(self.cmdapp.visible_prompt) :]]
            try:
                line = next(transcript)
            except StopIteration:
                line = ''
            line_num += 1
            # Read the entirety of a multi-line command
            while line.startswith(self.cmdapp.continuation_prompt):
                command_parts.append(line[len(self.cmdapp.continuation_prompt) :])
                try:
                    line = next(transcript)
                except StopIteration as exc:
                    msg = f'Transcript broke off while reading command beginning at line {line_num} with\n{command_parts[0]}'
                    raise StopIteration(msg) from exc
                line_num += 1
            command = ''.join(command_parts)
            # Send the command into the application and capture the resulting output
            stop = self.cmdapp.onecmd_plus_hooks(command)
            result = self.cmdapp.stdout.read()
            stop_msg = 'Command indicated application should quit, but more commands in transcript'
            # Read the expected result from transcript
            if su.strip_style(line).startswith(self.cmdapp.visible_prompt):
                message = f'\nFile {fname}, line {line_num}\nCommand was:\n{command}\nExpected: (nothing)\nGot:\n{result}\n'
                assert not result.strip(), message  # noqa: S101
                # If the command signaled the application to quit there should be no more commands
                assert not stop, stop_msg  # noqa: S101
                continue
            expected_parts = []
            while not su.strip_style(line).startswith(self.cmdapp.visible_prompt):
                expected_parts.append(line)
                try:
                    line = next(transcript)
                except StopIteration:
                    finished = True
                    break
                line_num += 1

            if stop:
                # This should only be hit if the command that set stop to True had output text
                assert finished, stop_msg  # noqa: S101

            # transform the expected text into a valid regular expression
            expected = ''.join(expected_parts)
            expected = self._transform_transcript_expected(expected)
            message = f'\nFile {fname}, line {line_num}\nCommand was:\n{command}\nExpected:\n{expected}\nGot:\n{result}\n'
            assert re.match(expected, result, re.MULTILINE | re.DOTALL), message  # noqa: S101

    def _transform_transcript_expected(self, s: str) -> str:
        r"""Parse the string with slashed regexes into a valid regex.

        Given a string like:

            Match a 10 digit phone number: /\d{3}-\d{3}-\d{4}/

        Turn it into a valid regular expression which matches the literal text
        of the string and the regular expression. We have to remove the slashes
        because they differentiate between plain text and a regular expression.
        Unless the slashes are escaped, in which case they are interpreted as
        plain text, or there is only one slash, which is treated as plain text
        also.

        Check the tests in tests/test_transcript.py to see all the edge
        cases.
        """
        regex = ''
        start = 0

        while True:
            (regex, first_slash_pos, start) = self._escaped_find(regex, s, start, False)
            if first_slash_pos == -1:
                # no more slashes, add the rest of the string and bail
                regex += re.escape(s[start:])
                break
            # there is a slash, add everything we have found so far
            # add stuff before the first slash as plain text
            regex += re.escape(s[start:first_slash_pos])
            start = first_slash_pos + 1
            # and go find the next one
            (regex, second_slash_pos, start) = self._escaped_find(regex, s, start, True)
            if second_slash_pos > 0:
                # add everything between the slashes (but not the slashes)
                # as a regular expression
                regex += s[start:second_slash_pos]
                # and change where we start looking for slashed on the
                # turn through the loop
                start = second_slash_pos + 1
            else:
                # No closing slash, we have to add the first slash,
                # and the rest of the text
                regex += re.escape(s[start - 1 :])
                break
        return regex

    @staticmethod
    def _escaped_find(regex: str, s: str, start: int, in_regex: bool) -> tuple[str, int, int]:
        """Find the next slash in {s} after {start} that is not preceded by a backslash.

        If we find an escaped slash, add everything up to and including it to regex,
        updating {start}. {start} therefore serves two purposes, tells us where to start
        looking for the next thing, and also tells us where in {s} we have already
        added things to {regex}

        {in_regex} specifies whether we are currently searching in a regex, we behave
        differently if we are or if we aren't.
        """
        while True:
            pos = s.find('/', start)
            if pos == -1:
                # no match, return to caller
                break
            if pos == 0:
                # slash at the beginning of the string, so it can't be
                # escaped. We found it.
                break
            # check if the slash is preceded by a backslash
            if s[pos - 1 : pos] == '\\':
                # it is.
                if in_regex:
                    # add everything up to the backslash as a
                    # regular expression
                    regex += s[start : pos - 1]
                    # skip the backslash, and add the slash
                    regex += s[pos]
                else:
                    # add everything up to the backslash as escaped
                    # plain text
                    regex += re.escape(s[start : pos - 1])
                    # and then add the slash as escaped
                    # plain text
                    regex += re.escape(s[pos])
                # update start to show we have handled everything
                # before it
                start = pos + 1
                # and continue to look
            else:
                # slash is not escaped, this is what we are looking for
                break
        return regex, pos, start



"""Supports the definition of commands in separate classes to be composed into cmd2.Cmd."""

from collections.abc import Callable, Mapping
from typing import (
    TYPE_CHECKING,
    TypeVar,
)

from .constants import (
    CLASS_ATTR_DEFAULT_HELP_CATEGORY,
    COMMAND_FUNC_PREFIX,
)
from .exceptions import (
    CommandSetRegistrationError,
)
from .utils import (
    Settable,
)

if TYPE_CHECKING:  # pragma: no cover
    import cmd2

#: Callable signature for a basic command  function
#: Further refinements are needed to define the input parameters
CommandFunc = Callable[..., bool | None]

CommandSetType = TypeVar('CommandSetType', bound=type['CommandSet'])


def with_default_category(category: str, *, heritable: bool = True) -> Callable[[CommandSetType], CommandSetType]:
    """Apply a category to all ``do_*`` command methods in a class that do not already have a category specified (Decorator).

    CommandSets that are decorated by this with `heritable` set to True (default) will set a class attribute that is
    inherited by all subclasses unless overridden. All commands of this CommandSet and all subclasses of this CommandSet
    that do not declare an explicit category will be placed in this category. Subclasses may use this decorator to
    override the default category.

    If `heritable` is set to False, then only the commands declared locally to this CommandSet will be placed in the
    specified category. Dynamically created commands and commands declared in sub-classes will not receive this
    category.

    :param category: category to put all uncategorized commands in
    :param heritable: Flag whether this default category should apply to sub-classes. Defaults to True
    :return: decorator function
    """

    def decorate_class(cls: CommandSetType) -> CommandSetType:
        if heritable:
            setattr(cls, CLASS_ATTR_DEFAULT_HELP_CATEGORY, category)

        import inspect

        from .constants import (
            CMD_ATTR_HELP_CATEGORY,
        )
        from .decorators import (
            with_category,
        )

        # get members of the class that meet the following criteria:
        # 1. Must be a function
        # 2. Must start with COMMAND_FUNC_PREFIX (do_)
        # 3. Must be a member of the class being decorated and not one inherited from a parent declaration
        methods = inspect.getmembers(
            cls,
            predicate=lambda meth: inspect.isfunction(meth)
            and meth.__name__.startswith(COMMAND_FUNC_PREFIX)
            and meth in inspect.getmro(cls)[0].__dict__.values(),
        )
        category_decorator = with_category(category)
        for method in methods:
            if not hasattr(method[1], CMD_ATTR_HELP_CATEGORY):
                setattr(cls, method[0], category_decorator(method[1]))
        return cls

    return decorate_class


class CommandSet:
    """Base class for defining sets of commands to load in cmd2.

    ``with_default_category`` can be used to apply a default category to all commands in the CommandSet.

    ``do_``, ``help_``, and ``complete_`` functions differ only in that self is the CommandSet instead of the cmd2 app
    """

    def __init__(self) -> None:
        """Private reference to the CLI instance in which this CommandSet running.

        This will be set when the CommandSet is registered and it should be
        accessed by child classes using the self._cmd property.
        """
        self.__cmd_internal: cmd2.Cmd | None = None

        self._settables: dict[str, Settable] = {}
        self._settable_prefix = self.__class__.__name__

    @property
    def _cmd(self) -> 'cmd2.Cmd':
        """Property for child classes to access self.__cmd_internal.

        Using this property ensures that self.__cmd_internal has been set
        and it tells type checkers that it's no longer a None type.

        Override this property to specify a more specific return type for static
        type checking. The typing.cast function can be used to assert to the
        type checker that the parent cmd2.Cmd instance is of a more specific
        subclass, enabling better autocompletion and type safety in the child class.

        For example:

            @property
            def _cmd(self) -> CustomCmdApp:
                return cast(CustomCmdApp, super()._cmd)


        :raises CommandSetRegistrationError: if CommandSet is not registered.
        """
        if self.__cmd_internal is None:
            raise CommandSetRegistrationError('This CommandSet is not registered')
        return self.__cmd_internal

    def on_register(self, cmd: 'cmd2.Cmd') -> None:
        """First step to registering a CommandSet, called by cmd2.Cmd.

        The commands defined in this class have not been added to the CLI object at this point.
        Subclasses can override this to perform any initialization requiring access to the Cmd object
        (e.g. configure commands and their parsers based on CLI state data).

        :param cmd: The cmd2 main application
        :raises CommandSetRegistrationError: if CommandSet is already registered.
        """
        if self.__cmd_internal is None:
            self.__cmd_internal = cmd
        else:
            raise CommandSetRegistrationError('This CommandSet has already been registered')

    def on_registered(self) -> None:
        """2nd step to registering, called by cmd2.Cmd after a CommandSet is registered and all its commands have been added.

        Subclasses can override this to perform custom steps related to the newly added commands (e.g. setting
        them to a disabled state).
        """

    def on_unregister(self) -> None:
        """First step to unregistering a CommandSet, called by ``cmd2.Cmd``.

        Subclasses can override this to perform any cleanup steps which require their commands being registered in the CLI.
        """

    def on_unregistered(self) -> None:
        """2nd step to unregistering, called by ``cmd2.Cmd`` after a CommandSet is unregistered and all its commands removed.

        Subclasses can override this to perform remaining cleanup steps.
        """
        self.__cmd_internal = None

    @property
    def settable_prefix(self) -> str:
        """Read-only accessor for the underlying private settable_prefix field."""
        return self._settable_prefix

    @property
    def settables(self) -> Mapping[str, Settable]:
        """Read-only accessor for the underlying private settables field."""
        return self._settables

    def add_settable(self, settable: Settable) -> None:
        """Add a settable parameter to the CommandSet.

        :param settable: Settable object being added
        """
        if self.__cmd_internal is not None:
            if not self._cmd.always_prefix_settables:
                if settable.name in self._cmd.settables and settable.name not in self._settables:
                    raise KeyError(f'Duplicate settable: {settable.name}')
            else:
                prefixed_name = f'{self._settable_prefix}.{settable.name}'
                if prefixed_name in self._cmd.settables and settable.name not in self._settables:
                    raise KeyError(f'Duplicate settable: {settable.name}')
        self._settables[settable.name] = settable

    def remove_settable(self, name: str) -> None:
        """Remove a settable parameter from the CommandSet.

        :param name: name of the settable being removed
        :raises KeyError: if the Settable matches this name
        """
        try:
            del self._settables[name]
        except KeyError as exc:
            raise KeyError(name + " is not a settable parameter") from exc

    def sigint_handler(self) -> bool:
        """Handle a SIGINT that occurred for a command in this CommandSet.

        :return: True if this completes the interrupt handling and no KeyboardInterrupt will be raised.
                 False to raise a KeyboardInterrupt.
        """
        return False



"""Module provides basic ability to copy from and paste to the clipboard/pastebuffer."""

import typing

import pyperclip  # type: ignore[import-untyped]


def get_paste_buffer() -> str:
    """Get the contents of the clipboard / paste buffer.

    :return: contents of the clipboard
    """
    return typing.cast(str, pyperclip.paste())


def write_to_paste_buffer(txt: str) -> None:
    """Copy text to the clipboard / paste buffer.

    :param txt: text to copy to the clipboard
    """
    pyperclip.copy(txt)



"""Import certain things for backwards compatibility."""

import contextlib
import importlib.metadata as importlib_metadata

with contextlib.suppress(importlib_metadata.PackageNotFoundError):
    __version__ = importlib_metadata.version(__name__)

from . import (
    plugin,
    rich_utils,
    string_utils,
)
from .argparse_completer import set_default_ap_completer_type
from .argparse_custom import (
    Cmd2ArgumentParser,
    Cmd2AttributeWrapper,
    CompletionItem,
    register_argparse_argument_parameter,
    set_default_argument_parser_type,
)
from .cmd2 import Cmd
from .colors import Color
from .command_definition import (
    CommandSet,
    with_default_category,
)
from .constants import (
    COMMAND_NAME,
    DEFAULT_SHORTCUTS,
)
from .decorators import (
    as_subcommand_to,
    with_argparser,
    with_argument_list,
    with_category,
)
from .exceptions import (
    Cmd2ArgparseError,
    CommandSetRegistrationError,
    CompletionError,
    PassThroughException,
    SkipPostcommandHooks,
)
from .parsing import Statement
from .py_bridge import CommandResult
from .rich_utils import RichPrintKwargs
from .string_utils import stylize
from .styles import Cmd2Style
from .utils import (
    CompletionMode,
    CustomCompletionSettings,
    Settable,
    categorize,
)

__all__: list[str] = [  # noqa: RUF022
    'COMMAND_NAME',
    'DEFAULT_SHORTCUTS',
    # Argparse Exports
    'Cmd2ArgumentParser',
    'Cmd2AttributeWrapper',
    'CompletionItem',
    'register_argparse_argument_parameter',
    'set_default_ap_completer_type',
    'set_default_argument_parser_type',
    # Cmd2
    'Cmd',
    'CommandResult',
    'CommandSet',
    'Statement',
    # Colors
    "Color",
    # Decorators
    'with_argument_list',
    'with_argparser',
    'with_category',
    'with_default_category',
    'as_subcommand_to',
    # Exceptions
    'Cmd2ArgparseError',
    'CommandSetRegistrationError',
    'CompletionError',
    'PassThroughException',
    'SkipPostcommandHooks',
    # modules
    'plugin',
    'rich_utils',
    'string_utils',
    # Rich Utils
    'RichPrintKwargs',
    # String Utils
    'stylize',
    # Styles,
    "Cmd2Style",
    # Utilities
    'categorize',
    'CompletionMode',
    'CustomCompletionSettings',
    'Settable',
]
